/*
Copyright (C) 1996-2001 Id Software, Inc.
Copyright (C) 2002-2009 John Fitzgibbons and others
Copyright (C) 2007-2008 Kristian Duske

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

// draw.c -- 2d drawing

#include "quakedef.h"

#define drawcolor_vp \
	"!!ARBvp1.0\n" \
	"TEMP scaled;\n" \
	"MAD scaled.x, vertex.attrib[0].x, program.env[0].x, program.env[0].z;\n" \
	"MAD scaled.y, vertex.attrib[0].y, program.env[0].y, program.env[0].w;\n" \
	"MOV scaled.z, vertex.attrib[0].z;\n" \
	"MOV scaled.w, vertex.attrib[0].w;\n" \
	TRANSFORMVERTEX ("result.position", "scaled") \
	"MOV result.color, vertex.attrib[1];\n" \
	"END\n"

#define drawcolor_fp \
	"!!ARBfp1.0\n" \
	"MOV result.color, fragment.color;\n" \
	"END\n"

GLuint arb_drawcolor_vp = 0;
GLuint arb_drawcolor_fp = 0;


#define drawtexture_vp \
	"!!ARBvp1.0\n" \
	"TEMP scaled;\n" \
	"MAD scaled.x, vertex.attrib[0].x, program.env[0].x, program.env[0].z;\n" \
	"MAD scaled.y, vertex.attrib[0].y, program.env[0].y, program.env[0].w;\n" \
	"MOV scaled.z, vertex.attrib[0].z;\n" \
	"MOV scaled.w, vertex.attrib[0].w;\n" \
	TRANSFORMVERTEX ("result.position", "scaled") \
	"MOV result.texcoord[0], vertex.attrib[2];\n" \
	"MOV result.color, vertex.attrib[1];\n" \
	"END\n"

#define drawtexture_fp \
	"!!ARBfp1.0\n" \
	"TEMP col1;\n" \
	"TEX col1, fragment.texcoord[0], texture[0], 2D;\n" \
	"MUL result.color, col1, fragment.color;\n" \
	"END\n"

GLuint arb_drawtexture_vp = 0;
GLuint arb_drawtexture_fp = 0;

#define MAX_DRAW_VERTS	4096

typedef struct drawvert_s
{
	float xyz[3];
	byte rgba[4];
	float st[2];
} drawvert_t;


void Draw_CreateShaders (void)
{
	arb_drawtexture_vp = GL_CreateProgram (GL_VERTEX_PROGRAM_ARB, drawtexture_vp, false);
	arb_drawtexture_fp = GL_CreateProgram (GL_FRAGMENT_PROGRAM_ARB, drawtexture_fp, false);

	arb_drawcolor_vp = GL_CreateProgram (GL_VERTEX_PROGRAM_ARB, drawcolor_vp, false);
	arb_drawcolor_fp = GL_CreateProgram (GL_FRAGMENT_PROGRAM_ARB, drawcolor_fp, false);
}


void Draw_Flush (void);

cvar_t		scr_conalpha = {"scr_conalpha", "0.5", true}; // johnfitz

qpic_t		*draw_disc;
qpic_t		*draw_backtile;

gltexture_t *char_texture; // johnfitz
qpic_t		*pic_ovr, *pic_ins; // johnfitz -- new cursor handling
qpic_t		*pic_nul; // johnfitz -- for missing gfx, don't crash

// johnfitz -- new pics
byte pic_ovr_data[8][8] =
{
	{255, 255, 255, 255, 255, 255, 255, 255},
	{255, 15, 15, 15, 15, 15, 15, 255},
	{255, 15, 15, 15, 15, 15, 15,  2},
	{255, 15, 15, 15, 15, 15, 15,  2},
	{255, 15, 15, 15, 15, 15, 15,  2},
	{255, 15, 15, 15, 15, 15, 15,  2},
	{255, 15, 15, 15, 15, 15, 15,  2},
	{255, 255,  2,  2,  2,  2,  2,  2},
};

byte pic_ins_data[9][8] =
{
	{ 15, 15, 255, 255, 255, 255, 255, 255},
	{ 15, 15,  2, 255, 255, 255, 255, 255},
	{ 15, 15,  2, 255, 255, 255, 255, 255},
	{ 15, 15,  2, 255, 255, 255, 255, 255},
	{ 15, 15,  2, 255, 255, 255, 255, 255},
	{ 15, 15,  2, 255, 255, 255, 255, 255},
	{ 15, 15,  2, 255, 255, 255, 255, 255},
	{ 15, 15,  2, 255, 255, 255, 255, 255},
	{255,  2,  2, 255, 255, 255, 255, 255},
};

byte pic_nul_data[8][8] =
{
	{252, 252, 252, 252,  0,  0,  0,  0},
	{252, 252, 252, 252,  0,  0,  0,  0},
	{252, 252, 252, 252,  0,  0,  0,  0},
	{252, 252, 252, 252,  0,  0,  0,  0},
	{  0,  0,  0,  0, 252, 252, 252, 252},
	{  0,  0,  0,  0, 252, 252, 252, 252},
	{  0,  0,  0,  0, 252, 252, 252, 252},
	{  0,  0,  0,  0, 252, 252, 252, 252},
};

byte pic_stipple_data[8][8] =
{
	{255,  0,  0,  0, 255,  0,  0,  0},
	{  0,  0, 255,  0,  0,  0, 255,  0},
	{255,  0,  0,  0, 255,  0,  0,  0},
	{  0,  0, 255,  0,  0,  0, 255,  0},
	{255,  0,  0,  0, 255,  0,  0,  0},
	{  0,  0, 255,  0,  0,  0, 255,  0},
	{255,  0,  0,  0, 255,  0,  0,  0},
	{  0,  0, 255,  0,  0,  0, 255,  0},
};

byte pic_crosshair_data[8][8] =
{
	{255, 255, 255, 255, 255, 255, 255, 255},
	{255, 255, 255,  8,  9, 255, 255, 255},
	{255, 255, 255,  6,  8,  2, 255, 255},
	{255,  6,  8,  8,  6,  8,  8, 255},
	{255, 255,  2,  8,  8,  2,  2,  2},
	{255, 255, 255,  7,  8,  2, 255, 255},
	{255, 255, 255, 255,  2,  2, 255, 255},
	{255, 255, 255, 255, 255, 255, 255, 255},
};
// johnfitz


byte pic_newcrosshair_data[16][16] =
{
	{0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff},
	{0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff},
	{0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xfe, 0xfe, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff},
	{0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xfe, 0xfe, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff},
	{0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xfe, 0xfe, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff},
	{0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff},
	{0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff},
	{0xff, 0xff, 0xfe, 0xfe, 0xfe, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xfe, 0xfe, 0xfe, 0xff, 0xff},
	{0xff, 0xff, 0xfe, 0xfe, 0xfe, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xfe, 0xfe, 0xfe, 0xff, 0xff},
	{0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff},
	{0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff},
	{0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xfe, 0xfe, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff},
	{0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xfe, 0xfe, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff},
	{0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xfe, 0xfe, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff},
	{0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff},
	{0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff},
};

gltexture_t *new_crosshair;


typedef struct
{
	gltexture_t *gltexture;
	float		sl, tl, sh, th;
} glpic_t;

ofs2dtype currentofs2d = OFS2D_NONE; // johnfitz -- for GL_Set2DOffsets

// ==============================================================================
//
//  PIC CACHING
//
// ==============================================================================

typedef struct cachepic_s
{
	char		name[MAX_QPATH];
	qpic_t		pic;
	byte		padding[32];	// for appended glpic
} cachepic_t;

#define	MAX_CACHED_PICS		128
cachepic_t	menu_cachepics[MAX_CACHED_PICS];
int			menu_numcachepics;

byte		menuplyr_pixels[4096];

//  scrap allocation
//  Allocate all the little status bar obejcts into a single texture
//  to crutch up stupid hardware / drivers

// mh - renamed to avoid the name #define in 2 different places with 2 different purposes
#define	MAX_SCRAPS		2
#define	SCRAP_BLOCK_WIDTH	256
#define	SCRAP_BLOCK_HEIGHT	256

int			scrap_allocated[MAX_SCRAPS][SCRAP_BLOCK_WIDTH];
byte		scrap_texels[MAX_SCRAPS][SCRAP_BLOCK_WIDTH * SCRAP_BLOCK_HEIGHT]; // johnfitz -- removed *4 after SCRAP_BLOCK_HEIGHT
qboolean	scrap_dirty;
gltexture_t	*scrap_textures[MAX_SCRAPS]; // johnfitz


/*
================
Scrap_AllocBlock

returns an index into scrap_texnums[] and the position inside it
================
*/
int Scrap_AllocBlock (int w, int h, int *x, int *y)
{
	int		i, j;
	int		best, best2;
	int		texnum;

	for (texnum = 0; texnum < MAX_SCRAPS; texnum++)
	{
		best = SCRAP_BLOCK_HEIGHT;

		for (i = 0; i < SCRAP_BLOCK_WIDTH - w; i++)
		{
			best2 = 0;

			for (j = 0; j < w; j++)
			{
				if (scrap_allocated[texnum][i+j] >= best)
					break;

				if (scrap_allocated[texnum][i+j] > best2)
					best2 = scrap_allocated[texnum][i+j];
			}

			if (j == w)
			{
				// this is a valid spot
				*x = i;
				*y = best = best2;
			}
		}

		if (best + h > SCRAP_BLOCK_HEIGHT)
			continue;

		for (i = 0; i < w; i++)
			scrap_allocated[texnum][*x + i] = best + h;

		return texnum;
	}

	Sys_Error ("Scrap_AllocBlock: full"); // johnfitz -- correct function name
	return 0; // johnfitz -- shut up compiler
}

/*
================
Scrap_Upload -- johnfitz -- now uses TexMgr
================
*/
void Scrap_Upload (void)
{
	char name[8];
	int	i;

	for (i = 0; i < MAX_SCRAPS; i++)
	{
		sprintf (name, "scrap%i", i);

		scrap_textures[i] = TM_LoadImage (NULL, name, SCRAP_BLOCK_WIDTH, SCRAP_BLOCK_HEIGHT, SRC_INDEXED, scrap_texels[i],
											  "", (src_offset_t) scrap_texels[i], TEXPREF_ALPHA | TEXPREF_OVERWRITE | TEXPREF_NOPICMIP | TEXPREF_CLAMP);
	}

	scrap_dirty = false;
}


/*
================
Draw_PicFromWad
================
*/
qpic_t *Draw_PicFromWad (char *name)
{
	qpic_t	*p;
	glpic_t	gl;
	src_offset_t offset; // johnfitz
	char texturename[64]; // johnfitz

	if (!(p = (qpic_t *) W_GetLumpName (name))) return pic_nul; // johnfitz

	// load little ones into the scrap
	if (p->width < 64 && p->height < 64)
	{
		int		x, y;
		int		i, j, k;
		int		texnum;

		texnum = Scrap_AllocBlock (p->width, p->height, &x, &y);
		scrap_dirty = true;
		k = 0;

		for (i = 0; i < p->height; i++)
		{
			for (j = 0; j < p->width; j++, k++)
				scrap_texels[texnum][(y+i) *SCRAP_BLOCK_WIDTH + x + j] = p->data[k];
		}

		gl.gltexture = scrap_textures[texnum]; // johnfitz -- changed to an array
		// johnfitz -- no longer go from 0.01 to 0.99
		gl.sl = x / (float) SCRAP_BLOCK_WIDTH;
		gl.sh = (x + p->width) / (float) SCRAP_BLOCK_WIDTH;
		gl.tl = y / (float) SCRAP_BLOCK_HEIGHT;
		gl.th = (y + p->height) / (float) SCRAP_BLOCK_HEIGHT;
	}
	else
	{
		sprintf (texturename, "%s:%s", WADFILENAME, name); // johnfitz

		offset = (src_offset_t) p - (src_offset_t) wad_base + sizeof (int) * 2; // johnfitz

		gl.gltexture = TM_LoadImage (NULL, texturename, p->width, p->height, SRC_INDEXED, p->data, WADFILENAME,
										 offset, TEXPREF_ALPHA | TEXPREF_PAD | TEXPREF_NOPICMIP); // johnfitz -- TexMgr

		gl.sl = 0;
		gl.sh = (float) p->width / (float) TM_PadConditional (p->width); // johnfitz
		gl.tl = 0;
		gl.th = (float) p->height / (float) TM_PadConditional (p->height); // johnfitz
	}

	memcpy (p->data, &gl, sizeof (glpic_t));

	return p;
}


/*
================
Draw_CachePic
================
*/
qpic_t	*Draw_CachePic (char *path)
{
	cachepic_t	*pic;
	int			i;
	qpic_t		*dat;
	glpic_t		gl;

	for (pic = menu_cachepics, i = 0; i < menu_numcachepics; pic++, i++)
	{
		if (!strcmp (path, pic->name))
			return &pic->pic;
	}

	Draw_Flush ();

	if (menu_numcachepics == MAX_CACHED_PICS)
		Sys_Error ("menu_numcachepics == MAX_CACHED_PICS");

	menu_numcachepics++;
	strcpy (pic->name, path);

	// load the pic from disk
	dat = (qpic_t *) COM_LoadTempFile (path);

	if (!dat)
	{
		dat = Draw_PicFromWad(path);
		/*return a 0-sized pic if we couldn't load it*/
		pic->pic.width = dat->width;
		pic->pic.height = dat->height;
		memcpy (pic->pic.data, dat->data, sizeof (glpic_t));
		return &pic->pic;
	}

	SwapPic (dat);

	// HACK HACK HACK --- we need to keep the bytes for
	// the translatable player picture just for the menu
	// configuration dialog
	if (!strcmp (path, "gfx/menuplyr.lmp"))
		memcpy (menuplyr_pixels, dat->data, dat->width * dat->height);

	pic->pic.width = dat->width;
	pic->pic.height = dat->height;

	gl.gltexture = TM_LoadImage (NULL, path, dat->width, dat->height, SRC_INDEXED, dat->data, path,
									 sizeof (int) * 2, TEXPREF_ALPHA | TEXPREF_PAD | TEXPREF_NOPICMIP); // johnfitz -- TexMgr

	gl.sl = 0;
	gl.sh = (float) dat->width / (float) TM_PadConditional (dat->width); // johnfitz
	gl.tl = 0;
	gl.th = (float) dat->height / (float) TM_PadConditional (dat->height); // johnfitz

	memcpy (pic->pic.data, &gl, sizeof (glpic_t));

	return &pic->pic;
}


/*
================
Draw_MakePic -- johnfitz -- generate pics from internal data
================
*/
qpic_t *Draw_MakePic (char *name, int width, int height, byte *data)
{
	int flags = TEXPREF_NEAREST | TEXPREF_ALPHA | TEXPREF_PERSIST | TEXPREF_NOPICMIP | TEXPREF_PAD;
	qpic_t		*pic;
	glpic_t		gl;

	pic = (qpic_t *) Hunk_Alloc (sizeof (qpic_t) - 4 + sizeof (glpic_t));
	pic->width = width;
	pic->height = height;

	gl.gltexture = TM_LoadImage (NULL, name, width, height, SRC_INDEXED, data, "", (src_offset_t) data, flags);
	gl.sl = 0;
	gl.sh = (float) width / (float) TM_PadConditional (width);
	gl.tl = 0;
	gl.th = (float) height / (float) TM_PadConditional (height);
	memcpy (pic->data, &gl, sizeof (glpic_t));

	return pic;
}

gltexture_t *Draw_MakeTexture (char *name, int width, int height, byte *data)
{
	int flags = TEXPREF_NEAREST | TEXPREF_ALPHA | TEXPREF_PERSIST | TEXPREF_NOPICMIP;
	return TM_LoadImage (NULL, name, width, height, SRC_INDEXED, data, "", (src_offset_t) data, flags);
}

// ==============================================================================
//
//  INIT
//
// ==============================================================================

/*
===============
Draw_LoadPics -- johnfitz
===============
*/
void Draw_LoadPics (void)
{
	byte		*data;
	src_offset_t	offset;

	data = (byte *) W_GetLumpName ("conchars");

	if (!data) Sys_Error ("Draw_LoadPics: couldn't load conchars");

	offset = (src_offset_t) data - (src_offset_t) wad_base;
	char_texture = TM_LoadImage (NULL, WADFILENAME":conchars", 128, 128, SRC_INDEXED, data,
									 WADFILENAME, offset, TEXPREF_ALPHA | TEXPREF_NEAREST | TEXPREF_NOPICMIP | TEXPREF_CONCHARS);

	draw_disc = Draw_PicFromWad ("disc");
	draw_backtile = Draw_PicFromWad ("backtile");
}


/*
===============
Draw_NewGame -- johnfitz
===============
*/
void Draw_NewGame (void)
{
	cachepic_t	*pic;
	int			i;

	// empty scrap and reallocate gltextures
	memset (&scrap_allocated, 0, sizeof (scrap_allocated));
	memset (&scrap_texels, 255, sizeof (scrap_texels));
	Scrap_Upload (); // creates 2 empty gltextures

	// empty lmp cache
	for (pic = menu_cachepics, i = 0; i < menu_numcachepics; pic++, i++)
		pic->name[0] = 0;

	menu_numcachepics = 0;

	// reload wad pics
	W_LoadWadFile (); // johnfitz -- filename is now hard-coded for honesty
	Draw_LoadPics ();
	SCR_LoadPics ();
	Sbar_LoadPics ();
}

/*
===============
Draw_Init -- johnfitz -- rewritten
===============
*/
void Draw_Init (void)
{
	Cvar_RegisterVariable (&scr_conalpha, NULL);

	// clear scrap and allocate gltextures
	memset (&scrap_allocated, 0, sizeof (scrap_allocated));
	memset (&scrap_texels, 255, sizeof (scrap_texels));
	Scrap_Upload (); // creates 2 empty textures

	// create internal pics
	pic_ins = Draw_MakePic ("ins", 8, 9, &pic_ins_data[0][0]);
	pic_ovr = Draw_MakePic ("ovr", 8, 8, &pic_ovr_data[0][0]);
	pic_nul = Draw_MakePic ("nul", 8, 8, &pic_nul_data[0][0]);

	// and the crosshair
	new_crosshair = Draw_MakeTexture ("crosshair", 16, 16, &pic_newcrosshair_data[0][0]);

	// load game pics
	Draw_LoadPics ();
}


// ==============================================================================
//
//  2D DRAWING
//
// ==============================================================================


float xofs_2d = 0;
float yofs_2d = 0;
float xscale_2d = 1;
float yscale_2d = 1;

drawvert_t *drawverts;
byte r_defaultdrawcolor[4] = {255, 255, 255, 255};

int draw_numvertexes = 0;

gltexture_t *draw_currenttexture = (gltexture_t *) 0xffffffff;

#define DM_COLORED		0
#define DM_TEXTURED		1
#define DM_NONE			2

int draw_currentmode = DM_NONE;
int draw_newmode = DM_NONE;


void Draw_Flush (void)
{
	// reset this first because we may be changing the array pointers
	drawverts = (drawvert_t *) scratchbuf;

	if (draw_newmode != draw_currentmode)
	{
		switch (draw_currentmode)
		{
		case DM_COLORED:
			GL_SetVertexProgram (arb_drawcolor_vp);
			GL_SetFragmentProgram (arb_drawcolor_fp);

			GL_EnableVertexArrays (VAA_0 | VAA_1);
			GL_VertexArrayPointer (0, 0, 3, GL_FLOAT, GL_FALSE, sizeof (drawvert_t), drawverts->xyz);
			GL_VertexArrayPointer (0, 1, 4, GL_UNSIGNED_BYTE, GL_TRUE, sizeof (drawvert_t), drawverts->rgba);
			break;

		case DM_TEXTURED:
		default:
			GL_SetVertexProgram (arb_drawtexture_vp);
			GL_SetFragmentProgram (arb_drawtexture_fp);

			GL_EnableVertexArrays (VAA_0 | VAA_1 | VAA_2);
			GL_VertexArrayPointer (0, 0, 3, GL_FLOAT, GL_FALSE, sizeof (drawvert_t), drawverts->xyz);
			GL_VertexArrayPointer (0, 1, 4, GL_UNSIGNED_BYTE, GL_TRUE, sizeof (drawvert_t), drawverts->rgba);
			GL_VertexArrayPointer (0, 2, 2, GL_FLOAT, GL_FALSE, sizeof (drawvert_t), drawverts->st);
			break;
		}

		draw_newmode = draw_currentmode;
	}

	if (draw_numvertexes)
	{
		GL_DrawArrays (GL_QUADS, 0, draw_numvertexes);
		draw_numvertexes = 0;
	}
}


void Draw_End (void)
{
	Draw_Flush ();
}


void Draw_EmitVertex (drawvert_t *vert, float x, float y, byte *rgba, float s, float t)
{
	vert->xyz[0] = x;
	vert->xyz[1] = y;
	vert->xyz[2] = 0;

	vert->rgba[0] = rgba[0];
	vert->rgba[1] = rgba[1];
	vert->rgba[2] = rgba[2];
	vert->rgba[3] = rgba[3];

	vert->st[0] = s;
	vert->st[1] = t;
}


void Draw_EmitQuad (int x, int y, int w, int h, byte *rgba, float sl, float tl, float sh, float th)
{
	Draw_EmitVertex (&drawverts[0], x, y, rgba, sl, tl);
	Draw_EmitVertex (&drawverts[1], x + w, y, rgba, sh, tl);
	Draw_EmitVertex (&drawverts[2], x + w, y + h, rgba, sh, th);
	Draw_EmitVertex (&drawverts[3], x, y + h, rgba, sl, th);

	draw_numvertexes += 4;
	drawverts += 4;
}


void Draw_SwitchState (int mode, gltexture_t *tex)
{
	if ((draw_numvertexes + 4) >= MAX_DRAW_VERTS) Draw_Flush ();

	if (mode != draw_currentmode)
	{
		Draw_Flush ();
		draw_currentmode = mode;
	}

	if (draw_currenttexture != tex)
	{
		Draw_Flush ();
		GL_BindTexture (GL_TEXTURE0, tex);
		draw_currenttexture = tex;
	}
}


void Draw_Colored (int x, int y, int w, int h, byte *rgba)
{
	Draw_SwitchState (DM_COLORED, draw_currenttexture);

	if (!rgba) rgba = r_defaultdrawcolor;

	Draw_EmitQuad (x, y, w, h, rgba, 0, 0, 1, 1);
}


void Draw_Textured (gltexture_t *tex, int x, int y, int w, int h, byte *rgba, float sl, float tl, float sh, float th)
{
	Draw_SwitchState (DM_TEXTURED, tex);

	if (!rgba) rgba = r_defaultdrawcolor;

	Draw_EmitQuad (x, y, w, h, rgba, sl, tl, sh, th);
}


void Draw_Textured2DQuad (float x, float y, float w, float h, float sl, float tl, float sh, float th, byte *color)
{
}


void Draw_Colored2DQuad (float x, float y, float w, float h, byte *color)
{
}


void Draw_DisableTexture (void)
{
}


void Draw_EnableTexture (void)
{
}


/*
================
Draw_CharacterQuad -- johnfitz -- seperate function to spit out verts
================
*/
void Draw_CharacterQuad (int x, int y, char num, byte *color)
{
	if (num != 32)
	{
		int				row, col;
		float			frow, fcol, size;

		row = num >> 4;
		col = num & 15;

		frow = row * 0.0625;
		fcol = col * 0.0625;
		size = 0.0625;

		Draw_Textured (char_texture, x, y, 8, 8, NULL, fcol, frow, fcol + size, frow + size);
	}
}


/*
================
Draw_Character -- johnfitz -- modified to call Draw_CharacterQuad
================
*/
void Draw_Character (int x, int y, int num)
{
	Draw_CharacterQuad (x, y, (num & 255), NULL);
}


void Draw_ColoredCharacter (int x, int y, int num, unsigned int color)
{
	Draw_CharacterQuad (x, y, (num & 255), (byte *) color);
}


/*
================
Draw_String -- johnfitz -- modified to call Draw_CharacterQuad
================
*/
void Draw_String (int x, int y, char *str)
{
	while (*str)
	{
		if (*str != 32) // don't waste verts on spaces
			Draw_CharacterQuad (x, y, ((*str) & 255), NULL);

		str++;
		x += 8;
	}
}


void Draw_Texture (int x, int y, int w, int h, void *tex, byte *color)
{
	Draw_Textured ((gltexture_t *) tex, x, y, w, h, color, 0, 0, 1, 1);
}


/*
=============
Draw_Pic -- johnfitz -- modified
=============
*/
void Draw_ColoredPic (int x, int y, qpic_t *pic, byte *color, int flags)
{
	glpic_t *gl;

	if (scrap_dirty)
	{
		Draw_Flush ();
		Scrap_Upload ();
		scrap_dirty = false;
	}

	gl = (glpic_t *) pic->data;

	if (flags & DP_MIRROR)
		Draw_Textured (gl->gltexture, x, y, pic->width, pic->height, color, gl->sh, gl->tl, gl->sl, gl->th);
	else Draw_Textured (gl->gltexture, x, y, pic->width, pic->height, color, gl->sl, gl->tl, gl->sh, gl->th);
}


void Draw_Pic (int x, int y, qpic_t *pic, int flags)
{
	Draw_ColoredPic (x, y, pic, NULL, flags);
}


/*
=============
Draw_TransPicTranslate -- johnfitz -- rewritten to use texmgr to do translation

Only used for the player color selection menu
=============
*/
void Draw_TransPicTranslate (int x, int y, qpic_t *pic, int top, int bottom)
{
	static int oldtop = -2;
	static int oldbottom = -2;
	gltexture_t *glt;
	glpic_t *glp;

	Draw_Flush ();

	if (top != oldtop || bottom != oldbottom)
	{
		oldtop = top;
		oldbottom = bottom;
		glp = (glpic_t *) pic->data;
		glt = glp->gltexture;
		TM_ReloadImage (glt, top, bottom);
	}

	Draw_Pic (x, y, pic, DP_DEFAULT);
}


/*
================
Draw_ConsoleBackground -- johnfitz -- rewritten
================
*/
void Draw_ConsoleBackground (void)
{
	qpic_t *pic;
	float alpha;

	pic = Draw_CachePic ("gfx/conback.lmp");
	pic->width = vid.conwidth;
	pic->height = vid.conheight;

	alpha = (con_forcedup) ? 1.0 : scr_conalpha.value;

	// 	GL_Set2DOffsets (OFS2D_CONSOLE); // in case this is called from weird places

	if (alpha > 0.0)
	{
		byte color[4] = {255, 255, 255, 255};
		color[3] = BYTE_CLAMPF (alpha);

		Draw_ColoredPic (0, 0, pic, color, DP_DEFAULT);
	}
}


/*
=============
Draw_TileClear

This repeats a 64*64 tile graphic to fill the screen around a sized down
refresh window.
=============
*/
void Draw_TileClear (int x, int y, int w, int h)
{
	Draw_Textured (((glpic_t *) draw_backtile->data)->gltexture, x, y, w, h, NULL, x / 64.0, y / 64.0, (x + w) / 64.0, (y + h) / 64.0);
}


/*
=============
Draw_Fill

Fills a box of pixels with a single color
=============
*/
void Draw_Fill (int x, int y, int w, int h, int c, float alpha) // johnfitz -- added alpha
{
	// fitzquake bug - this used to overwrite the alpha value in the palette table
	unsigned int color = d_8to24table_rgba[c];
	byte *rgba = (byte *) &color;

	rgba[3] = BYTE_CLAMPF (alpha);
	Draw_Colored (x, y, w, h, rgba);
}


/*
================
Draw_FadeScreen -- johnfitz -- revised
================
*/
void Draw_FadeScreen (void)
{
	byte fadecolor[4] = {0, 0, 0, 160};

	GL_Set2DOffsets (OFS2D_DEFAULT);
	Draw_Colored (0, 0, vid.glwidth, vid.glheight, fadecolor);
}


/*
================
Draw_BeginDisc

Draws the little blue disc in the corner of the screen.
Call before beginning any disc IO.
================
*/
void Draw_BeginDisc (void)
{
	// mh - direct writing to the front buffer is not actually supported on most modern gfx chips!
}


/*
================
GL_Set2DOffsets

================
*/
void GL_Set2DOffsets (ofs2dtype newofs2d)
{
	float s;
	int lines;

	if (newofs2d == currentofs2d)
		return;

	Draw_Flush ();
	currentofs2d = newofs2d;

	// enforce defaults
	xofs_2d = 0;
	yofs_2d = 0;
	xscale_2d = 1;
	yscale_2d = 1;

	switch (newofs2d)
	{
	case OFS2D_CONSOLE:
		lines = vid.conheight - ((scr_con_current * vid.conheight) / vid.glheight);

		xofs_2d = 0;
		yofs_2d = scr_con_current - vid.glheight;

		xscale_2d = (float) vid.glwidth / (float) vid.conwidth;
		yscale_2d = (float) vid.glheight / (float) vid.conheight;
		break;

	case OFS2D_MENU:
		s = min ((float) vid.glwidth / 320.0, (float) vid.glheight / 200.0);
		s = CLAMP (1.0, scr_menuscale.value, s);

		xofs_2d = (vid.glwidth - (320 * s)) / 2;
		yofs_2d = (vid.glheight - (200 * s)) / 2;

		xscale_2d = s;
		yscale_2d = s;
		break;

	case OFS2D_SBAR:
		s = CLAMP (1.0, scr_sbarscale.value, (float) vid.glwidth / 320.0);

		if (cl.gametype == GAME_DEATHMATCH)
			xofs_2d = 0;
		else xofs_2d = (vid.glwidth - (320.0f * s)) / 2;

		switch (vid.sblines)
		{
		case 24:
			yofs_2d = vid.glheight - 48;
			break;

		default:
			yofs_2d = vid.glheight - vid.sblines;
			break;
		}

		xscale_2d = s;
		yscale_2d = s;

		break;

	case OFS2D_BOTTOMLEFT: // used by devstats
		s = (float) vid.glwidth / (float) vid.conwidth; // use console scale

		xofs_2d = 0;
		yofs_2d = vid.glheight - (200.0f * s);

		xscale_2d = s;
		yscale_2d = s;

		break;

	case OFS2D_BOTTOMRIGHT: // used by fps/clock
		s = (float) vid.glwidth / (float) vid.conwidth; // use console scale

		xofs_2d = vid.glwidth - (320.0f * s);
		yofs_2d = vid.glheight - (200.0f * s);

		xscale_2d = s;
		yscale_2d = s;

		break;

	case OFS2D_DEFAULT:
	default:
		// just use default values from above
		break;
	}

	// setup offsets for console/etc scaling (yayyy, we're using the GPU now!)
	qglProgramEnvParameter4fARB (GL_VERTEX_PROGRAM_ARB, 0, xscale_2d, yscale_2d, xofs_2d, yofs_2d);
}


/*
================
GL_Set2D -- johnfitz -- rewritten
================
*/
void GL_Set2D (void)
{
	currentofs2d = -1;

	glViewport (0, 0, vid.glwidth, vid.glheight);

	glMatrixMode (GL_PROJECTION);
    glLoadIdentity ();
	glOrtho (0, vid.glwidth, vid.glheight, 0, -99999, 99999);

	glMatrixMode (GL_MODELVIEW);
    glLoadIdentity ();

	glDisable (GL_DEPTH_TEST);
	glDepthMask (GL_FALSE);
	glDisable (GL_CULL_FACE);

	glEnable (GL_BLEND);
	glDisable (GL_ALPHA_TEST);

	// cached environment for draw batching
	draw_numvertexes = 0;
	draw_currenttexture = (gltexture_t *) 0xffffffff;
	draw_currentmode = DM_NONE;
	draw_newmode = DM_NONE;

	drawverts = (drawvert_t *) scratchbuf;

	GL_Set2DOffsets (OFS2D_DEFAULT);
}

