/*
Copyright (C) 1996-2001 Id Software, Inc.
Copyright (C) 2002-2009 John Fitzgibbons and others

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
// r_main.c

#include "quakedef.h"

#define underwater_vp \
	"!!ARBvp1.0\n" \
	"TEMP tc0;\n" \
	"MOV result.position, vertex.attrib[0];\n" \
	"MAD tc0, vertex.attrib[2], 2.0, program.local[0].x;\n" \
	"MUL result.texcoord[1], tc0, 0.024543;\n" \
	"MOV result.texcoord[0], vertex.attrib[2];\n" \
	"MOV result.color, vertex.attrib[1];\n" \
	"END\n"

#define underwater_fp \
	"!!ARBfp1.0\n" \
	"TEMP tc0, tc1;\n" \
	"SIN tc0.x, fragment.texcoord[1].y;\n" \
	"SIN tc0.y, fragment.texcoord[1].x;\n" \
	"MAD tc1, tc0, program.local[0].z, fragment.texcoord[0];\n" \
	"MUL tc0, tc1, program.local[0].y;\n" \
	"TEX tc1, tc0, texture[0], 2D;\n" \
	"LRP result.color, fragment.color.a, fragment.color, tc1;\n" \
	"END\n"

GLuint arb_underwater_vp = 0;
GLuint arb_underwater_fp = 0;


#define polyblend_vp \
	"!!ARBvp1.0\n" \
	"MOV result.position, vertex.attrib[0];\n" \
	"MOV result.color, program.local[0];\n" \
	"END\n"

#define bbox_vp \
	"!!ARBvp1.0\n" \
	TRANSFORMVERTEX ("result.position", "vertex.attrib[0]") \
	"MOV result.color, program.local[0];\n" \
	"END\n"

#define polyblend_fp \
	"!!ARBfp1.0\n" \
	"MOV result.color, fragment.color;\n" \
	"END\n"

#define showtris_fp \
	"!!ARBfp1.0\n" \
	"MOV result.color, {0.666, 0.666, 0.666, 1};\n" \
	"END\n"

GLuint arb_bbox_vp = 0;
GLuint arb_polyblend_vp = 0;
GLuint arb_polyblend_fp = 0;
GLuint arb_showtris_fp = 0;


#define blur_vp \
	"!!ARBvp1.0\n" \
	"TEMP tc0;\n" \
	"MOV result.position, vertex.attrib[0];\n" \
	"MOV result.color, vertex.attrib[1];\n" \
	"MOV result.texcoord[0], vertex.attrib[2];\n" \
	"MOV result.texcoord[1], program.local[0];\n" \
	"MOV result.texcoord[2], vertex.attrib[0];\n" \
	"END\n"

#define blur_fp \
	"!!ARBfp1.0\n" \
	"TEMP col0;\n" \
	"DP3 col0, fragment.texcoord[2], fragment.texcoord[2];\n" \
	"MUL col0, col0, program.local[0].w;\n" \
	"MIN col0, col0, 0.95;\n" \
	"TEMP tex0;\n" \
	"TEX tex0, fragment.texcoord[0], texture[0], 2D;\n" \
	"MUL tex0, tex0, fragment.color;\n" \
	"MUL tex0.w, tex0.w, col0.w;\n" \
	"LRP result.color, fragment.texcoord[1].a, fragment.texcoord[1], tex0;\n" \
	"END\n"


GLuint arb_blur_vp = 0;
GLuint arb_blur_fp = 0;


void Main_CreateShaders (void)
{
	arb_underwater_vp = GL_CreateProgram (GL_VERTEX_PROGRAM_ARB, underwater_vp, false);
	arb_underwater_fp = GL_CreateProgram (GL_FRAGMENT_PROGRAM_ARB, underwater_fp, false);

	arb_blur_vp = GL_CreateProgram (GL_VERTEX_PROGRAM_ARB, blur_vp, false);
	arb_blur_fp = GL_CreateProgram (GL_FRAGMENT_PROGRAM_ARB, blur_fp, false);

	arb_bbox_vp = GL_CreateProgram (GL_VERTEX_PROGRAM_ARB, bbox_vp, false);
	arb_polyblend_vp = GL_CreateProgram (GL_VERTEX_PROGRAM_ARB, polyblend_vp, false);
	arb_polyblend_fp = GL_CreateProgram (GL_FRAGMENT_PROGRAM_ARB, polyblend_fp, false);
	arb_showtris_fp = GL_CreateProgram (GL_FRAGMENT_PROGRAM_ARB, showtris_fp, false);
}


cvar_t r_colorfilter = {"r_colorfilter", "7"};

void R_DrawCoronas (void);

glmatrix r_world_matrix;

void R_SetupAliasFrame (entity_t *e, aliashdr_t *paliashdr, int frame, aliasstate_t *state);
void R_SetupIQMFrame (entity_t *e, iqmdata_t *iqm, int frame, aliasstate_t *state);
void GLARB_SetupAliasLerp (entity_t *e, aliashdr_t *hdr);
void GLARB_LerpAliasMove (entity_t *e, aliasstate_t *state);

qboolean showtris = false;

void R_BBoxForEnt (entity_t *e)
{
	int i, j;
	float mins[3];
	float maxs[3];
	vec3_t bbox[8];
	vec3_t fv, rv, uv;
	float angles[3];

	if (!e->model) return;

	// figure the base bbox size
	if (e->model->type == mod_alias)
	{
		aliasstate_t *state = &e->aliasstate;
		aliashdr_t *hdr = (aliashdr_t *) Mod_Extradata (e->model);
		aliasbbox_t *bboxes = (aliasbbox_t *) ((byte *) hdr + hdr->bboxes);

		// wipe the state and set up interpolation
		//memset (&e->aliasstate, 0, sizeof (aliasstate_t));
		GLARB_SetupAliasLerp (e, hdr);
		GLARB_LerpAliasMove (e, state);

		for (i = 0; i < 3; i++)
		{
			mins[i] = bboxes[e->previouspose].mins[i] * (1.0f - state->blend) + bboxes[e->currentpose].mins[i] * state->blend;
			maxs[i] = bboxes[e->previouspose].maxs[i] * (1.0f - state->blend) + bboxes[e->currentpose].maxs[i] * state->blend;
		}
	}
	else if (e->model->type == mod_iqm)
	{
		aliasstate_t *state = &e->aliasstate;

		// wipe the state and set up interpolation
		memset (&e->aliasstate, 0, sizeof (aliasstate_t));
		R_SetupIQMFrame (e, e->model->iqmdata, e->frame, state);
		GLARB_LerpAliasMove (e, state);

		// we can probably do better than this but we don't have per-pose bounding boxes yet and any culling is better than none
		VectorCopy (e->model->mins, mins);
		VectorCopy (e->model->mins, maxs);
	}
	else if (e->model->type == mod_brush)
	{
		VectorCopy (e->model->bbmins, mins);
		VectorCopy (e->model->bbmaxs, maxs);
	}
	else
	{
		VectorCopy (e->model->mins, mins);
		VectorCopy (e->model->maxs, maxs);
	}

	// compute a full bounding box
	for (i = 0; i < 8; i++)
	{
		// the bounding box is expanded by 1 unit in each direction so
		// that it won't z-fight with the model (if it's a tight box)
		bbox[i][0] = (i & 1) ? mins[0] - 1.0f : maxs[0] + 1.0f;
		bbox[i][1] = (i & 2) ? mins[1] - 1.0f : maxs[1] + 1.0f;
		bbox[i][2] = (i & 4) ? mins[2] - 1.0f : maxs[2] + 1.0f;
	}

	// these factors hold valid for both MDLs and brush models; tested brush models with rmq rotate test
	// and ne_tower; tested alias models by assigning bobjrotate to angles 0/1/2 and observing the result
	// i guess that ID just left out angles[2] because it never really happened in the original game
	if (e->model->type == mod_brush)
	{
		angles[0] = -e->angles[0];
		angles[1] = -e->angles[1];
		angles[2] = -e->angles[2];
	}
	else
	{
		angles[0] = e->angles[0];
		angles[1] = -e->angles[1];
		angles[2] = -e->angles[2];
	}

	// derive forward/right/up vectors from the angles
	AngleVectors (angles, fv, rv, uv);

	// compute the rotated bbox corners
	mins[0] = mins[1] = mins[2] = 9999999;
	maxs[0] = maxs[1] = maxs[2] = -9999999;

	// and rotate the bounding box
	for (i = 0; i < 8; i++)
	{
		vec3_t tmp;

		VectorCopy (bbox[i], tmp);

		bbox[i][0] = DotProduct (fv, tmp);
		bbox[i][1] = -DotProduct (rv, tmp);
		bbox[i][2] = DotProduct (uv, tmp);

		// and convert them to mins and maxs
		for (j = 0; j < 3; j++)
		{
			if (bbox[i][j] < mins[j]) mins[j] = bbox[i][j];
			if (bbox[i][j] > maxs[j]) maxs[j] = bbox[i][j];
		}
	}

	// translate the bbox to it's final position at the entity origin
	VectorAdd (e->origin, mins, e->mins);
	VectorAdd (e->origin, maxs, e->maxs);

	if (e->scale != 16)
	{
		// scale the bbox
		VectorScale (e->mins, (float) e->scale / 16.0f, e->mins);
		VectorScale (e->maxs, (float) e->scale / 16.0f, e->maxs);
	}

	// true origin of entity is at bbox center point (needed for bmodels
	// where the origin could be at (0, 0, 0) or at a corner)
	e->trueorigin[0] = e->mins[0] + (e->maxs[0] - e->mins[0]) * 0.5f;
	e->trueorigin[1] = e->mins[1] + (e->maxs[1] - e->mins[1]) * 0.5f;
	e->trueorigin[2] = e->mins[2] + (e->maxs[2] - e->mins[2]) * 0.5f;
}


void R_AddVisEdict (entity_t *e)
{
	// add to the visedicts list
	if (cl_numvisedicts < MAX_VISEDICTS)
	{
		cl_visedicts[cl_numvisedicts] = e;
		cl_numvisedicts++;

		// figure the entities bounding box with proper correction for rotation
		R_BBoxForEnt (e);

		// mark as visible (static entities need this so that they won't be added twice)
		e->visframe = r_framecount;
	}
}


void R_DrawIQMModels (void);
void R_DrawAliasModels (void);
void R_DrawSpriteModels (void);
void R_PrepDLights (void);

void R_ShowTrisBegin (void)
{
	Fog_DisableGlobalFog ();

	if (r_showtris.value < 2)
	{
		glDisable (GL_DEPTH_TEST);
	}
	else
	{
		glEnable (GL_POLYGON_OFFSET_FILL);
		glEnable (GL_POLYGON_OFFSET_LINE);
		glPolygonOffset (-1, -3);
	}

	glDepthMask (GL_FALSE);
	glPolygonMode (GL_FRONT_AND_BACK, GL_LINE);
	showtris = true;
}


void R_ShowTrisEnd (void)
{
	glPolygonMode (GL_FRONT_AND_BACK, GL_FILL);

	if (r_showtris.value < 2)
	{
		glEnable (GL_DEPTH_TEST);
	}
	else
	{
		glDisable (GL_POLYGON_OFFSET_FILL);
		glDisable (GL_POLYGON_OFFSET_LINE);
	}

	glDepthMask (GL_TRUE);
	Fog_EnableGlobalFog ();
	showtris = false;
}


qboolean	r_cache_thrash;		// compatability

vec3_t		r_entorigin;

int			r_visframecount;	// bumped when going to a new PVS
int			r_framecount;		// used for dlight push checking

glmatrix r_projection_matrix;
mplane_t	frustum[4];

// johnfitz -- rendering statistics
int rs_brushpolys, rs_aliaspolys;
int rs_drawelements;
int rs_dynamiclightmaps;

qboolean	envmap;				// true during envmap command capture

//
// view origin
//
vec3_t	vup;
vec3_t	vpn;
vec3_t	vright;
vec3_t	r_origin;

float r_fovx, r_fovy; // johnfitz -- rendering fov may be different becuase of r_waterwarp

//
// screen size info
//
refdef_t	r_refdef;

mleaf_t		*r_viewleaf, *r_oldviewleaf;

float		d_lightstylevalue[256];	// 8.8 fraction of base light value


cvar_t	r_norefresh = {"r_norefresh", "0"};
cvar_t	r_drawentities = {"r_drawentities", "1"};
cvar_t	r_drawviewmodel = {"r_drawviewmodel", "1"};
cvar_t	r_speeds = {"r_speeds", "0"};
cvar_t	r_fullbright = {"r_fullbright", "0"};
cvar_t	r_lightmap = {"r_lightmap", "0"};
cvar_t	r_shadows = {"r_shadows", "0", true};
cvar_t	r_wateralpha = {"r_wateralpha", "1", true};
cvar_t	r_dynamic = {"r_dynamic", "1", true};
cvar_t	r_novis = {"r_novis", "0", true};

cvar_t	gl_finish = {"gl_finish", "0"};
cvar_t	gl_clear = {"gl_clear", "0"};
cvar_t	gl_cull = {"gl_cull", "1"};
cvar_t	gl_smoothmodels = {"gl_smoothmodels", "1"};
cvar_t	gl_affinemodels = {"gl_affinemodels", "0"};
cvar_t	gl_polyblend = {"gl_polyblend", "1"};
cvar_t	gl_playermip = {"gl_playermip", "0"};
cvar_t	gl_nocolors = {"gl_nocolors", "0"};

// johnfitz -- new cvars
cvar_t	r_clearcolor = {"r_clearcolor", "2", true};
cvar_t	r_drawflat = {"r_drawflat", "0"};
cvar_t	r_flatlightstyles = {"r_flatlightstyles", "0"};
cvar_t	gl_fullbrights = {"gl_fullbrights", "1", true};
cvar_t	r_oldskyleaf = {"r_oldskyleaf", "0"};
cvar_t	r_drawworld = {"r_drawworld", "1"};
cvar_t	r_showbboxes = {"r_showbboxes", "0"};
cvar_t	r_lerpmodels = {"r_lerpmodels", "1"};
cvar_t	r_lerpmove = {"r_lerpmove", "1"};
cvar_t	r_nolerp_list = {"r_nolerp_list", "progs/flame.mdl,progs/flame2.mdl,progs/braztall.mdl,progs/brazshrt.mdl,progs/longtrch.mdl,progs/flame_pyre.mdl,progs/v_saw.mdl,progs/v_xfist.mdl,progs/h2stuff/newfire.mdl"};

extern cvar_t	r_vfog;
// johnfitz

cvar_t r_motionblur = {"r_motionblur", "1", true};
cvar_t r_motionblurradius = {"r_motionblurradius", "5", true};
cvar_t r_motionblurfade = {"r_motionblurfade", "1", true};
cvar_t r_motionblurupdatetime = {"r_motionblurupdatetime", "1", true};

cvar_t r_lockpvs = {"r_lockpvs", "0"};
cvar_t r_lockfrustum = {"r_lockfrustum", "0"};

qboolean r_dowarp = false;

void R_DrawViewModel (void);
void R_AlphaListBeginFrame (void);
void R_DrawAlphaList (void);

/*
=================
R_CullSphere

Returns true if the sphere is outside the frustum
grabbed from AlienArena 2011 source (repaying the compliment!!!)
=================
*/
qboolean R_CullSphere (const float *center, const float radius, const int clipflags)
{
	int		i;
	mplane_t *p;

	for (i = 0, p = frustum; i < 4; i++, p++)
	{
		if (!(clipflags & (1 << i)))
			continue;

		if (DotProduct (center, p->normal) - p->dist <= -radius)
			return true;
	}

	return false;
}


/*
=================
R_CullBox -- johnfitz -- replaced with new function from lordhavoc

Returns true if the box is completely outside the frustum
=================
*/
qboolean R_CullBox (vec3_t emins, vec3_t emaxs)
{
	int i;
	mplane_t *p;

	for (i = 0; i < 4; i++)
	{
		p = &frustum[i];

		switch (p->signbits)
		{
		default:
		case 0:
			if (p->normal[0] * emaxs[0] + p->normal[1] * emaxs[1] + p->normal[2] * emaxs[2] < p->dist)
				return true;

			break;

		case 1:
			if (p->normal[0] * emins[0] + p->normal[1] * emaxs[1] + p->normal[2] * emaxs[2] < p->dist)
				return true;

			break;

		case 2:
			if (p->normal[0] * emaxs[0] + p->normal[1] * emins[1] + p->normal[2] * emaxs[2] < p->dist)
				return true;

			break;

		case 3:
			if (p->normal[0] * emins[0] + p->normal[1] * emins[1] + p->normal[2] * emaxs[2] < p->dist)
				return true;

			break;

		case 4:
			if (p->normal[0] * emaxs[0] + p->normal[1] * emaxs[1] + p->normal[2] * emins[2] < p->dist)
				return true;

			break;

		case 5:
			if (p->normal[0] * emins[0] + p->normal[1] * emaxs[1] + p->normal[2] * emins[2] < p->dist)
				return true;

			break;

		case 6:
			if (p->normal[0] * emaxs[0] + p->normal[1] * emins[1] + p->normal[2] * emins[2] < p->dist)
				return true;

			break;

		case 7:
			if (p->normal[0] * emins[0] + p->normal[1] * emins[1] + p->normal[2] * emins[2] < p->dist)
				return true;

			break;
		}
	}

	return false;
}


// ==============================================================================
//
// SETUP FRAME
//
// ==============================================================================

int SignbitsForPlane (mplane_t *out)
{
	int	bits, j;

	// for fast box on planeside test
	bits = 0;

	for (j = 0; j < 3; j++)
	{
		if (out->normal[j] < 0)
			bits |= 1 << j;
	}

	return bits;
}


/*
=============
GL_SetFrustum
=============
*/
void GL_SetFrustum (glmatrix *matrix, float fovx, float fovy)
{
#define NEARCLIP 4
	float left, right, bottom, top;
	float nudge = 1.0 - 1.0 / (1 << 23);

	right = NEARCLIP * tan (fovx * M_PI / 360.0);
	left = -right;

	top = NEARCLIP * tan (fovy * M_PI / 360.0);
	bottom = -top;

	matrix->m16[0] = (2 * NEARCLIP) / (right - left);
	matrix->m16[4] = 0;
	matrix->m16[8] = (right + left) / (right - left);
	matrix->m16[12] = 0;

	matrix->m16[1] = 0;
	matrix->m16[5] = (2 * NEARCLIP) / (top - bottom);
	matrix->m16[9] = (top + bottom) / (top - bottom);
	matrix->m16[13] = 0;

	matrix->m16[2] = 0;
	matrix->m16[6] = 0;
	matrix->m16[10] = -1 * nudge;
	matrix->m16[14] = -2 * NEARCLIP * nudge;

	matrix->m16[3] = 0;
	matrix->m16[7] = 0;
	matrix->m16[11] = -1;
	matrix->m16[15] = 0;

	glLoadMatrixf (matrix->m16);
}


void R_ExtractFrustum (void)
{
	int i;
	glmatrix r_mvp_matrix;

	// get the frustum from the actual matrixes that were used
	GL_MultiplyMatrix (&r_mvp_matrix, &r_world_matrix, &r_projection_matrix);

	// frustum 0
	frustum[0].normal[0] = r_mvp_matrix.m16[3] - r_mvp_matrix.m16[0];
	frustum[0].normal[1] = r_mvp_matrix.m16[7] - r_mvp_matrix.m16[4];
	frustum[0].normal[2] = r_mvp_matrix.m16[11] - r_mvp_matrix.m16[8];

	// frustum 1
	frustum[1].normal[0] = r_mvp_matrix.m16[3] + r_mvp_matrix.m16[0];
	frustum[1].normal[1] = r_mvp_matrix.m16[7] + r_mvp_matrix.m16[4];
	frustum[1].normal[2] = r_mvp_matrix.m16[11] + r_mvp_matrix.m16[8];

	// frustum 2
	frustum[2].normal[0] = r_mvp_matrix.m16[3] + r_mvp_matrix.m16[1];
	frustum[2].normal[1] = r_mvp_matrix.m16[7] + r_mvp_matrix.m16[5];
	frustum[2].normal[2] = r_mvp_matrix.m16[11] + r_mvp_matrix.m16[9];

	// frustum 3
	frustum[3].normal[0] = r_mvp_matrix.m16[3] - r_mvp_matrix.m16[1];
	frustum[3].normal[1] = r_mvp_matrix.m16[7] - r_mvp_matrix.m16[5];
	frustum[3].normal[2] = r_mvp_matrix.m16[11] - r_mvp_matrix.m16[9];

	for (i = 0; i < 4; i++)
	{
		VectorNormalize (frustum[i].normal);

		frustum[i].type = PLANE_ANYZ;
		frustum[i].dist = DotProduct (r_origin, frustum[i].normal); // FIXME: shouldn't this always be zero?
		frustum[i].signbits = SignbitsForPlane (&frustum[i]);
	}
}


// the motion blur texture needs to be separate because it doesn't get refreshed every frame
gltexture_t gl_postproc_texture;
gltexture_t gl_motionblur_texture;

int gl_rtt_width = 0;
int gl_rtt_height = 0;

void R_CreateRTTTexture (char *name, gltexture_t *tex)
{
	int hunkmark = Hunk_LowMark ();

	glGenTextures (1, &tex->texnum);
	GL_BindTexture (GL_DONT_CARE, tex);
	glTexImage2D (GL_TEXTURE_2D, 0, GL_RGBA8, gl_rtt_width, gl_rtt_height, 0, GL_BGRA, GL_UNSIGNED_INT_8_8_8_8_REV, Hunk_Alloc (gl_rtt_width * gl_rtt_height * 4));

	glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);

	Con_Printf ("Created %s at %ix%i\n", name, gl_rtt_width, gl_rtt_height);

	Hunk_FreeToLowMark (hunkmark);
}


void R_CreateRTTTextures (int width, int height)
{
	int maxrtt;

	// figure the size to create the texture at
	glGetIntegerv (GL_MAX_TEXTURE_SIZE, &maxrtt);

	// scale to powers of 2
	for (gl_rtt_width = 1; gl_rtt_width < width; gl_rtt_width <<= 1);
	for (gl_rtt_height = 1; gl_rtt_height < height; gl_rtt_height <<= 1);

	// clamp to max texture size
	if (gl_rtt_width > maxrtt) gl_rtt_width = maxrtt;
	if (gl_rtt_height > maxrtt) gl_rtt_height = maxrtt;

	// create the textures for RTT effects
	R_CreateRTTTexture ("gl_postproc_texture", &gl_postproc_texture);
	R_CreateRTTTexture ("gl_motionblur_texture", &gl_motionblur_texture);
}


/*
=============
R_SetupGL
=============
*/
int gl_vpx, gl_vpy, gl_vpwidth, gl_vpheight;

void R_SetupGL (void)
{
	// evaluate viewport size
	if (r_refdef.vrectset)
	{
		gl_vpx = r_refdef.vrect.x;
		gl_vpy = r_refdef.vrect.y;
		gl_vpwidth = r_refdef.vrect.width;
		gl_vpheight = r_refdef.vrect.height;
	}
	else
	{
		gl_vpx = 0;
		gl_vpy = vid.sblines;
		gl_vpwidth = vid.glwidth;
		gl_vpheight = vid.glheight - vid.sblines;
	}

	if (r_dowarp)
	{
		// bound viewport dimensions to the rtt texture size
		if (gl_vpwidth > gl_rtt_width) gl_vpwidth = gl_rtt_width;
		if (gl_vpheight > gl_rtt_height) gl_vpheight = gl_rtt_height;
	}
}


vec3_t r_oldorigin;
vec3_t r_oldangles;
extern qboolean r_rebuildworld;

int cached = 0;
int uncached = 0;


void R_SetViewport (void)
{
	// now set the correct viewport
	glViewport (gl_vpx, gl_vpy, gl_vpwidth, gl_vpheight);

	if (memcmp (r_oldorigin, r_refdef.vieworg, sizeof (vec3_t)) || memcmp (r_oldangles, r_refdef.viewangles, sizeof (vec3_t)))
	{
		memcpy (r_oldorigin, r_refdef.vieworg, sizeof (vec3_t));
		memcpy (r_oldangles, r_refdef.viewangles, sizeof (vec3_t));
		r_rebuildworld = true;
		uncached++;
	}
	else cached++;

	// Con_Printf ("cached surfs %i\n", (cached * 100) / (cached + uncached));

	glMatrixMode (GL_PROJECTION);
	GL_SetFrustum (&r_projection_matrix, r_fovx, r_fovy); // johnfitz -- use r_fov* vars

	glMatrixMode (GL_MODELVIEW);

	GL_IdentityMatrix (&r_world_matrix);

	GL_RotateMatrix (&r_world_matrix, -90, 1, 0, 0);
	GL_RotateMatrix (&r_world_matrix, 90, 0, 0, 1);

	GL_RotateMatrix (&r_world_matrix, -r_refdef.viewangles[2], 1, 0, 0);
	GL_RotateMatrix (&r_world_matrix, -r_refdef.viewangles[0], 0, 1, 0);
	GL_RotateMatrix (&r_world_matrix, -r_refdef.viewangles[1], 0, 0, 1);

	GL_TranslateMatrix (&r_world_matrix, -r_refdef.vieworg[0], -r_refdef.vieworg[1], -r_refdef.vieworg[2]);

	glLoadMatrixf (r_world_matrix.m16);

	R_ExtractFrustum ();

	// set drawing parms
	if (gl_cull.value)
		glEnable (GL_CULL_FACE);
	else glDisable (GL_CULL_FACE);

	glDisable (GL_BLEND);
	glDisable (GL_ALPHA_TEST);
	glEnable (GL_DEPTH_TEST);
	glDepthMask (GL_TRUE);
	glDepthFunc (GL_LEQUAL);
}


/*
=============
R_Clear -- johnfitz -- rewritten and gutted
=============
*/
void R_SetClearColor_f (void);

void R_Clear (void)
{
	extern int gl_stencilbits;
	unsigned int clearbits;

	glDepthRange (0, 1);

	// always clear the depth buffer
	clearbits = GL_DEPTH_BUFFER_BIT;

	if (r_viewleaf && r_viewleaf->contents == CONTENTS_SOLID)
	{
		// when noclipping set a black clear color
		glClearColor (0.15, 0.15, 0.15, 1);
		clearbits |= GL_COLOR_BUFFER_BIT;
	}
	else if (r_colorfilter.value != 7)
	{
		glClearColor (0, 0, 0, 1);
		clearbits |= GL_COLOR_BUFFER_BIT;
	}
	else if (r_lockpvs.value || r_lockfrustum.value)
	{
		glClearColor (1, 0.5, 0, 1);
		clearbits |= GL_COLOR_BUFFER_BIT;
	}
	else if (gl_clear.value)
	{
		// set the correct clear color
		R_SetClearColor_f ();
		clearbits |= GL_COLOR_BUFFER_BIT;
	}

	// if we got a stencil buffer we must always clear it, even if we're not using it
	if (gl_stencilbits) clearbits |= GL_STENCIL_BUFFER_BIT;

	glClear (clearbits);

	if (r_colorfilter.value != 7)
	{
		int cf = (int) r_colorfilter.value & 7;
		int r = 0, g = 0, b = 0;

		if (cf & 1) r = 1;
		if (cf & 2) g = 1;
		if (cf & 4) b = 1;

		glColorMask (r, g, b, 1);
	}
}

/*
===============
R_SetupScene -- johnfitz -- this is the stuff that needs to be done once per eye in stereo mode
===============
*/
void R_SetupScene (void)
{
	R_AnimateLight ();
	r_framecount++;
}


void R_UpdateContentsColor (void)
{
	/*
	if (r_viewleaf->contents == CONTENTS_EMPTY)
	{
		// brought to you today by "disgusting hacks 'r' us"...
		extern cshift_t cshift_empty;

		cl.cshifts[CSHIFT_CONTENTS].destcolor[0] = cshift_empty.destcolor[0];
		cl.cshifts[CSHIFT_CONTENTS].destcolor[1] = cshift_empty.destcolor[1];
		cl.cshifts[CSHIFT_CONTENTS].destcolor[2] = cshift_empty.destcolor[2];
		cl.cshifts[CSHIFT_CONTENTS].percent = cshift_empty.percent;
	}
	else if (r_viewleaf->contentscolor)
	{
		// if the viewleaf has a contents colour set we just override with it
		cl.cshifts[CSHIFT_CONTENTS].destcolor[0] = r_viewleaf->contentscolor[0];
		cl.cshifts[CSHIFT_CONTENTS].destcolor[1] = r_viewleaf->contentscolor[1];
		cl.cshifts[CSHIFT_CONTENTS].destcolor[2] = r_viewleaf->contentscolor[2];

		// these now have more colour so reduce the percent to compensate
		if (r_viewleaf->contents == CONTENTS_WATER)
			cl.cshifts[CSHIFT_CONTENTS].percent = 48;
		else if (r_viewleaf->contents == CONTENTS_LAVA)
			cl.cshifts[CSHIFT_CONTENTS].percent = 112;
		else if (r_viewleaf->contents == CONTENTS_SLIME)
			cl.cshifts[CSHIFT_CONTENTS].percent = 80;
		else cl.cshifts[CSHIFT_CONTENTS].percent = 0;
	}
	else
	*/
	{
		// empty or undefined
		V_SetContentsColor (r_viewleaf->contents);
	}

	// now calc the blend
	V_CalcBlend ();
}


/*
===============
R_SetupView -- johnfitz -- this is the stuff that needs to be done once per frame, even in stereo mode
===============
*/
void R_SetupView (void)
{
	Fog_SetupFrame (); // johnfitz

	// build the transformation matrix for the given view angles
	VectorCopy (r_refdef.vieworg, r_origin);
	AngleVectors (r_refdef.viewangles, vpn, vright, vup);

	// current viewleaf
	r_oldviewleaf = r_viewleaf;
	r_viewleaf = Mod_PointInLeaf (r_origin, cl.worldmodel);

	// mh - replaced with new version
	R_UpdateContentsColor ();
	// V_SetContentsColor (r_viewleaf->contents);
	// V_CalcBlend ();

	r_cache_thrash = false;

	// johnfitz -- calculate r_fovx and r_fovy here
	r_fovx = r_refdef.fov_x;
	r_fovy = r_refdef.fov_y;

	// (this is the same condition as software Quake used so it's a valid baseline to work from; we add cl.inwater
	// for func_water support but expect some latency between when the server sets it and the client recieves it)
	r_dowarp = r_waterwarp.value && ((r_viewleaf->contents <= CONTENTS_WATER) || (cl.cshifts[CSHIFT_CONTENTS].percent > 0));// || cl.inwater);

	R_MarkLeafs (); // johnfitz -- create texture chains from PVS

	// removed bogus client-side cheat-protection
}

// ==============================================================================
//
// RENDER VIEW
//
// ==============================================================================

/*
================
R_RenderScene
================
*/
cvar_t r_uwfactor = {"r_uwfactor", "512", true};
cvar_t r_uwscale = {"r_uwscale", "2", true};

void RTTWarp_EmitVertex (float *vert, float x, float y, float *rgba, float s, float t)
{
	vert[0] = x;
	vert[1] = y;
	vert[2] = 0;

	vert[3] = rgba[0];
	vert[4] = rgba[1];
	vert[5] = rgba[2];
	vert[6] = rgba[3];

	vert[7] = s;
	vert[8] = t;
}


void R_DrawUnderwaterWarp (void)
{
	float params[4] = {0, 0, 0, 0};

	float sh, th;
	float xl, yl, xh, yh;

	float *verts = (float *) scratchbuf;

	if (!r_dowarp) return;
	if (!gl_postproc_texture.texnum) return;

	// this will be replaced by our shader warp...
	glViewport (0, 0, vid.glwidth, vid.glheight);

	glDepthMask (GL_FALSE);
	glDisable (GL_ALPHA_TEST);
	glDisable (GL_BLEND);
	glDisable (GL_DEPTH_TEST);
	glDisable (GL_CULL_FACE);

	GL_SetVertexProgram (arb_underwater_vp);
	GL_SetFragmentProgram (arb_underwater_fp);

	params[0] = cl.ctime * 10.18591625f * 4.0f;
	params[1] = 1.0f / r_uwfactor.value;
	params[2] = r_uwscale.value;

	qglProgramLocalParameter4fvARB (GL_VERTEX_PROGRAM_ARB, 0, params);
	qglProgramLocalParameter4fvARB (GL_FRAGMENT_PROGRAM_ARB, 0, params);

	GL_BindTexture (GL_TEXTURE0, &gl_postproc_texture);
	glCopyTexSubImage2D (GL_TEXTURE_2D, 0, 0, 0, 0, vid.sblines, vid.glwidth, vid.glheight - vid.sblines);

	sh = ((float) vid.glwidth / (float) gl_rtt_width) * r_uwfactor.value;
	th = ((float) (vid.glheight - vid.sblines) / (float) gl_rtt_height) * r_uwfactor.value;

	GL_EnableVertexArrays (VAA_0 | VAA_1 | VAA_2);
	GL_VertexArrayPointer (0, 0, 3, GL_FLOAT, GL_FALSE, sizeof (float) * 9, &verts[0]);
	GL_VertexArrayPointer (0, 1, 4, GL_FLOAT, GL_FALSE, sizeof (float) * 9, &verts[3]);
	GL_VertexArrayPointer (0, 2, 2, GL_FLOAT, GL_FALSE, sizeof (float) * 9, &verts[7]);

	// expand the verts to overcome warp distortion
#define WARP_EXPAND 8.0f

	xl = (-WARP_EXPAND / (vid.glwidth / 2.0f)) - 1.0f;
	xh = ((vid.glwidth + WARP_EXPAND) / (vid.glwidth / 2.0f)) - 1.0f;
	yl = ((vid.sblines - WARP_EXPAND) / (vid.glheight / 2.0f)) - 1.0f;
	yh = ((vid.glheight + WARP_EXPAND) / (vid.glheight / 2.0f)) - 1.0f;

	// get polyblend for free here
	RTTWarp_EmitVertex (&verts[0], xl, yl, v_blend, 0, 0);
	RTTWarp_EmitVertex (&verts[9], xh, yl, v_blend, sh, 0);
	RTTWarp_EmitVertex (&verts[18], xh, yh, v_blend, sh, th);
	RTTWarp_EmitVertex (&verts[27], xl, yh, v_blend, 0, th);

	GL_DrawArrays (GL_QUADS, 0, 4);

	// disable view blending
	v_blend[3] = -1;
}


void GL_PolyBlend (void)
{
	float blendverts[] = {-1, -1, 0, 1, -1, 0, 1, 1, 0, -1, 1, 0};

	// whaddaya know - these run faster
	if (!gl_polyblend.value) return;
	if (v_blend[3] <= 0) return;
	if (r_dowarp) return;

	// fixme - need to implement some state filtering here...
	glDepthMask (GL_FALSE);
	glDisable (GL_ALPHA_TEST);
	glEnable (GL_BLEND);
	glDisable (GL_DEPTH_TEST);
	glDisable (GL_CULL_FACE);

	glViewport (0, 0, vid.glwidth, vid.glheight);

	GL_EnableVertexArrays (VAA_0);
	GL_VertexArrayPointer (0, 0, 3, GL_FLOAT, GL_FALSE, 0, blendverts);

	GL_SetVertexProgram (arb_polyblend_vp);
	GL_SetFragmentProgram (arb_polyblend_fp);
	qglProgramLocalParameter4fvARB (GL_VERTEX_PROGRAM_ARB, 0, v_blend);

	GL_DrawArrays (GL_QUADS, 0, 4);
}


void RTTBlur_EmitVertex (float *vert, float x, float y, float *rgba, float s, float t)
{
	vert[0] = x;
	vert[1] = y;
	vert[2] = 0;

	vert[3] = rgba[0];
	vert[4] = rgba[1];
	vert[5] = rgba[2];
	vert[6] = rgba[3];

	vert[7] = s;
	vert[8] = t;
}


void R_DrawMotionBlur (void)
{
	static int blurstate = 0;
	float accblur;
	static float blurtime = 0.0f;

	float color[4] = {1, 1, 1, 0.75f};
	float *verts = (float *) scratchbuf;

	if (!gl_motionblur_texture.texnum) return;

	if (!(r_motionblur.value > 0.0f))
	{
		blurtime = 0.0f;
		return;
	}

	// evaluate blur conditions
	if (cl.stats[STAT_HEALTH] <= 0)
	{
		// being dead always overrides everything else
		accblur = 0.99f;
	}
	else if (cl.cshifts[CSHIFT_DAMAGE].percent > 0)
	{
		// initial damage blur
		accblur = cl.damageblur; //0.5f;
		// Con_Printf ("blur at %f\n", accblur);
	}
	else if (cl.damageblur > 0.0f)
	{
		// persist and fade damage blur
		cl.damageblur -= (cl.ctime - cl.oldtime) * r_motionblurfade.value;
		accblur = cl.damageblur; //0.5f;
	}
	else accblur = -1.0f;

	if (accblur <= 0.0f)
	{
		// reinit if we're not blurring so that the contents of the
		// accumulation buffer are valid for the frame
		blurstate = 0;
		blurtime = 0.0f;
		return;
	}

	if (!blurstate)
	{
		// the first entry just refreshes the blur texture
		GL_BindTexture (GL_TEXTURE0, &gl_motionblur_texture);
		glCopyTexSubImage2D (GL_TEXTURE_2D, 0, 0, 0, 0, 0, vid.glwidth, vid.glheight);
		blurstate = 1;
		return;
	}

	color[3] = accblur * r_motionblur.value;

	// above 1 the scene will be fully overwritten by the blur texture; clamp to this so that we at least see something...
	if (color[3] > 0.95f) color[3] = 0.95f;

	glViewport (0, 0, vid.glwidth, vid.glheight);

	glDepthMask (GL_FALSE);
	glDisable (GL_ALPHA_TEST);
	glEnable (GL_BLEND);
	glDisable (GL_DEPTH_TEST);
	glDisable (GL_CULL_FACE);

	GL_SetVertexProgram (arb_blur_vp);
	GL_SetFragmentProgram (arb_blur_fp);

	if (v_blend[3] > 0)
		qglProgramLocalParameter4fvARB (GL_VERTEX_PROGRAM_ARB, 0, v_blend);
	else qglProgramLocalParameter4fARB (GL_VERTEX_PROGRAM_ARB, 0, 1, 1, 1, 0);

	qglProgramLocalParameter4fARB (GL_FRAGMENT_PROGRAM_ARB, 0, 0, 0, 0, r_motionblurradius.value + sin (cl.ctime * 5.0f));

	GL_BindTexture (GL_TEXTURE0, &gl_motionblur_texture);

	// blend over scene
	GL_EnableVertexArrays (VAA_0 | VAA_1 | VAA_2);
	GL_VertexArrayPointer (0, 0, 3, GL_FLOAT, GL_FALSE, sizeof (float) * 9, &verts[0]);
	GL_VertexArrayPointer (0, 1, 4, GL_FLOAT, GL_FALSE, sizeof (float) * 9, &verts[3]);
	GL_VertexArrayPointer (0, 2, 2, GL_FLOAT, GL_FALSE, sizeof (float) * 9, &verts[7]);

	// build verts (reusing the underwater warp setup)
	RTTBlur_EmitVertex (&verts[0], -1, -1, color, 0, 0);
	RTTBlur_EmitVertex (&verts[9], 1, -1, color, (float) vid.glwidth / (float) gl_rtt_width, 0);
	RTTBlur_EmitVertex (&verts[18], 1, 1, color, (float) vid.glwidth / (float) gl_rtt_width, (float) vid.glheight / (float) gl_rtt_height);
	RTTBlur_EmitVertex (&verts[27], -1, 1, color, 0, (float) vid.glheight / (float) gl_rtt_height);

	GL_DrawArrays (GL_QUADS, 0, 4);

	if (blurtime < cl.ctime)
	{
		// refresh the blur texture at constant rate to prevent fps dependence
		glCopyTexSubImage2D (GL_TEXTURE_2D, 0, 0, 0, 0, 0, vid.glwidth, vid.glheight);
		blurtime = cl.ctime + (0.025f * r_motionblurupdatetime.value);
	}

	// disable view blending
	v_blend[3] = -1;
}


void R_EmitWirePoint (vec3_t origin)
{
	int size = 8;

	glBegin (GL_LINES);

	glVertex3f (origin[0] - size, origin[1], origin[2]);
	glVertex3f (origin[0] + size, origin[1], origin[2]);
	glVertex3f (origin[0], origin[1] - size, origin[2]);
	glVertex3f (origin[0], origin[1] + size, origin[2]);
	glVertex3f (origin[0], origin[1], origin[2] - size);
	glVertex3f (origin[0], origin[1], origin[2] + size);

	glEnd ();
}


void R_EmitWireBox (vec3_t mins, vec3_t maxs)
{
	glBegin (GL_QUAD_STRIP);

	glVertex3f (mins[0], mins[1], mins[2]);
	glVertex3f (mins[0], mins[1], maxs[2]);
	glVertex3f (maxs[0], mins[1], mins[2]);
	glVertex3f (maxs[0], mins[1], maxs[2]);
	glVertex3f (maxs[0], maxs[1], mins[2]);
	glVertex3f (maxs[0], maxs[1], maxs[2]);
	glVertex3f (mins[0], maxs[1], mins[2]);
	glVertex3f (mins[0], maxs[1], maxs[2]);
	glVertex3f (mins[0], mins[1], mins[2]);
	glVertex3f (mins[0], mins[1], maxs[2]);

	glEnd ();
}


void R_ShowBBoxes2 (void)
{
	int i;
	entity_t *ent;
	qboolean stateset = false;

	for (i = 0; i < cl_numvisedicts; i++)
	{
		ent = cl_visedicts[i];

		if (!stateset)
		{
			// disable vertex arrays because we're going to immediate mode
			GL_EnableVertexArrays (0);

			glPolygonMode (GL_FRONT_AND_BACK, GL_LINE);
			glDisable (GL_CULL_FACE);

			GL_SetVertexProgram (arb_bbox_vp);
			GL_SetFragmentProgram (arb_polyblend_fp);
			qglProgramLocalParameter4fARB (GL_VERTEX_PROGRAM_ARB, 0, 0.666, 0.666, 0.666, 1);

			stateset = true;
		}

		R_EmitWireBox (ent->mins, ent->maxs);
	}

	if (stateset)
	{
		glEnable (GL_CULL_FACE);
		glPolygonMode (GL_FRONT_AND_BACK, GL_FILL);
	}
}


void R_ShowBBoxes (void)
{
	int i;
	edict_t *ed;
	extern edict_t *sv_player;
	vec3_t mins, maxs;
	qboolean stateset = false;

	if (!r_showbboxes.value) return;
	if (cl.maxclients > 1) return;
	if (!sv.active) return;
	if (!r_drawentities.value) return;

	for (i = 1; i < pr_progs->num_edicts; i++)
	{
		ed = pr_progs->edicts[i];

		// don't draw player's own bbox
		if (ed == sv_player) continue;

		// don't show free edicts
		if (ed->free) continue;

		// show the actual bboxes that will be used for world interaction
		VectorCopy (ed->v.absmin, mins);
		VectorCopy (ed->v.absmax, maxs);

 		if (R_CullBox (mins, maxs)) continue;

		if (!stateset)
		{
			// disable vertex arrays because we're going to immediate mode
			GL_EnableVertexArrays (0);

 			glDisable (GL_DEPTH_TEST);
			glPolygonMode (GL_FRONT_AND_BACK, GL_LINE);
			glDisable (GL_CULL_FACE);

			GL_SetVertexProgram (arb_bbox_vp);
			GL_SetFragmentProgram (arb_polyblend_fp);
			qglProgramLocalParameter4fARB (GL_VERTEX_PROGRAM_ARB, 0, 0.666, 0.666, 0.666, 1);

			stateset = true;
		}

		if (ed->v.mins[0] == ed->v.maxs[0] && ed->v.mins[1] == ed->v.maxs[1] && ed->v.mins[2] == ed->v.maxs[2])
		{
			// point entity
			R_EmitWirePoint (ed->v.origin);
		}
		else
		{
			// box entity
			R_EmitWireBox (mins, maxs);
		}
	}

	if (stateset)
	{
		glEnable (GL_CULL_FACE);
		glPolygonMode (GL_FRONT_AND_BACK, GL_FILL);
 		glEnable (GL_DEPTH_TEST);
	}
}


void R_RenderSceneObjects (void)
{
	R_AlphaListBeginFrame ();
	R_DrawWorld ();
	R_DrawAliasModels ();
	R_DrawIQMModels ();
	R_DrawSpriteModels ();
	R_DrawParticles ();
	R_DrawCoronas ();
	R_DrawAlphaList ();
	R_DrawViewModel ();
}


void R_RenderScene (void)
{
	// mh 2010-07-16
	R_SetupScene (); // johnfitz -- this does everything that should be done once per call to RenderScene
	R_SetupGL ();
	R_SetViewport ();
	R_Clear ();

	Fog_EnableGlobalFog ();
	R_RenderSceneObjects ();
	Fog_DisableGlobalFog ();

	if (r_showbboxes.value > 1)
		R_ShowBBoxes2 ();
	else R_ShowBBoxes ();

	if (r_showtris.value > 0)
	{
		R_ShowTrisBegin ();
		R_RenderSceneObjects ();
		R_ShowTrisEnd ();
	}
}


/*
================
R_RenderView
================
*/
void R_RenderView (void)
{
	int time1 = 0, time2 = 0;

	// for splitting from V_RenderView
	if (con_forcedup) return;
	if (!cl_entities[cl.viewentity]) return;

	if (r_norefresh.value) return;
	if (!cl.worldmodel) Sys_Error ("R_RenderView: NULL worldmodel");
	if (r_speeds.value) time1 = SDL_GetTicks ();

	// johnfitz -- rendering statistics
	// mh - always so that we can use con_printf for debug output while developing
	rs_brushpolys = rs_aliaspolys = rs_drawelements = rs_dynamiclightmaps = 0;

	R_PrepDLights ();	// precalc some dlight properties once per frame (instead of once per surface) for better runtime perf
	R_SetupView (); // johnfitz -- this does everything that should be done once per frame
	R_RenderScene ();

	if (r_colorfilter.value != 7)
	{
		glColorMask (1, 1, 1, 1);
	}

	R_DrawUnderwaterWarp ();

	// run motion blur
	R_DrawMotionBlur ();

	// moved polyblend back to here because it's faster
	// polyblend needs to come after any RTT effects because they can switch off the blend
	GL_PolyBlend ();

	// cleaned up code and made it much more readable here
	if (r_speeds.value)
	{
		// johnfitz -- modified r_speeds output
		// mh - moved this inside the "if"
		time2 = SDL_GetTicks ();

		Con_Printf ("%3i ms  %4i wpoly   %4i epoly   %4i draw   %3i lmap\n",
			time2 - time1,
			rs_brushpolys,
			rs_aliaspolys,
			rs_drawelements,
			rs_dynamiclightmaps);
	}
	// johnfitz
}


